@@ -1,6 +1,6 @@
 ;;; emms-tag-editor.el --- Edit track tags. -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2006-2021 Free Software Foundation, Inc.
+;; Copyright (C) 2006-2023 Free Software Foundation, Inc.
 ;;
 ;; Original Author: Ye Wenbin <wenbinye@163.com>
 ;; Authors: the Emms developers (see AUTHORS file)
@@ -37,6 +37,8 @@
 (require 'emms-info-mp3info)
 (require 'emms-playlist-mode)
 (require 'emms-mark)
+(require 'emms-cache)
+(require 'emms-tag-tracktag)
 (require 'format-spec)
 (require 'subr-x)
 
@@ -67,24 +69,25 @@
 (defun emms-tag-editor-make-format (tags)
   "Make a format string based on TAGS."
   (concat "%m\n" (emms-propertize (format "%-16s = " "name")
-                             'read-only t 'rear-nonsticky t
-                             'face 'bold)
+				  'read-only t 'rear-nonsticky t
+				  'face 'bold)
           "%f\n"
           (mapconcat
            (lambda (tag)
              (concat (emms-propertize (format "%-16s = " (symbol-name tag))
-                                 'read-only t 'rear-nonsticky t
-                                 'face 'bold)
+                                      'read-only t 'rear-nonsticky t
+                                      'face 'bold)
                      "%" (cdr (assoc tag emms-tag-editor-tags))))
            tags "\n")
           "\n\n"))
 
 (defvar emms-tag-editor-formats
-  (let* ((tags (mapcar 'car emms-tag-editor-tags))
+  (let* ((tags (mapcar #'car emms-tag-editor-tags))
          (default (emms-tag-editor-make-format (remove 'info-date tags))))
     `(("mp3" . ,default)
       ("ogg" . ,(emms-tag-editor-make-format (remove 'info-year tags)))
       ("flac" . ,(emms-tag-editor-make-format (remove 'info-year tags)))
+      ("opus" . ,(emms-tag-editor-make-format (remove 'info-genre tags)))
       ("default" . ,default)))
   "Format to use when inserting the track.
 The CAR part is the extension of the track name, and the CDR part
@@ -115,14 +118,14 @@ the track, and returns a string to insert in
 used for inserting the track.")
 
 (defvar emms-tag-editor-parse-function 'emms-tag-editor-default-parser
-  "Function to parse tags in `emms-tag-editor-edit-buffer'.
+  "Function to parse tags in `emms-tag-editor-edit-buffer\\='.
 It should find all modified tags, and return all the tracks.  The
 tracks for which a tag has been modified should set a property
-'tag-modified to t.  If the track name has been changed, the
-function should set a new property 'newname instead of setting
-the 'name directly.
+\\='tag-modified to t.  If the track name has been changed, the
+function should set a new property \\='newname instead of setting
+the \\='name directly.
 
-See also `emms-tag-editor-default-parser'.")
+See also `emms-tag-editor-default-parser\\='.")
 
 (defvar emms-tag-editor-tagfile-functions
   '(("mp3" "mid3v2"
@@ -138,7 +141,8 @@ See also `emms-tag-editor-default-parser'.")
       (info-performer   . "--TOPE")
       (info-date        . "--TDAT")))
     ("ogg" . emms-tag-editor-tag-ogg)
-    ("flac" . emms-tag-editor-tag-flac))
+    ("flac" . emms-tag-editor-tag-flac)
+    ("opus" . emms-tag-tracktag-file))
   "An alist used when committing changes to tags in files.
 If the external program sets tags by command line options
 one-by-one, then the list should like:
@@ -153,7 +157,7 @@ See also `emms-tag-editor-tag-file' and `emms-tag-editor-tag-ogg'.")
   "Commit changes to an FLAC file according to TRACK."
   (require 'emms-info-metaflac)
   (with-temp-buffer
-    (let ((tags '("artist" "composer" "performer" "title" "album" "tracknumber" "discnumber" "date" "genre" "note"))
+    (let ((tags '("albumartist" "artist" "composer" "performer" "title" "album" "tracknumber" "discnumber" "date" "genre" "note"))
 	  val)
       (mapc (lambda (tag)
               (let ((info-tag (intern (concat "info-" tag))))
@@ -201,7 +205,7 @@ Valid tags are given by `emms-tag-editor-tagfile-functions'."
             (if (and val (stringp val))
                 (setq args (append (list (cdr tag) val) args))))
           tags)
-    (apply 'call-process program
+    (apply #'call-process program
            nil (get-buffer-create emms-tag-editor-log-buffer) nil
            (nconc args (list filename)))))
 
@@ -222,12 +226,12 @@ This string is suitable for inserting into the tags buffer."
         (funcall format track)
       (format-spec
        format
-       (apply 'format-spec-make
+       (apply #'format-spec-make
               ?m (emms-propertize (emms-track-force-description track)
                                   'face 'emms-playlist-track-face
                                   'emms-track (copy-sequence track))
               ?f (emms-track-name track)
-              (apply 'append
+              (apply #'append
                      (mapcar (lambda (tag)
                                (list (string-to-char (cdr tag))
                                      (or (emms-track-get track (car tag)) "")))
@@ -265,49 +269,75 @@ This string is suitable for inserting into the tags buffer."
     (emms-tag-editor-erase-buffer emms-tag-editor-log-buffer)
     (emms-tag-editor-erase-buffer emms-tag-editor-edit-buffer)
     (set-buffer (get-buffer emms-tag-editor-edit-buffer))
-    (mapc 'emms-tag-editor-insert-track tracks)
+    (mapc #'emms-tag-editor-insert-track tracks)
     (emms-tag-editor-mode)
     (pop-to-buffer (current-buffer))
     (goto-char (point-min))
     (emms-tag-editor-display-log-buffer-maybe)))
 
-(defun emms-tag-editor-edit-track (track)
-  "Edit the track at point, or TRACK."
-  (interactive (list (emms-tag-editor-track-at)))
-  (if (null track)
-      (message "No track at point!")
-    (emms-tag-editor-insert-tracks (list track))))
+(defun emms-tag-editor--tagfile-function (track)
+  "Return value of `emms-tag-editor-tagfile-functions' for TRACK, or nil."
+  (assoc (file-name-extension (emms-track-get track 'name))
+         emms-tag-editor-tagfile-functions))
 
-(defun emms-tag-editor-edit-marked-tracks ()
-  "Edit all tracks marked in the current buffer."
+(defun emms-tag-editor--track-editable-p (track)
+  "Return t if TRACK is not a file, or has a tagfile function defined."
+  (or (not (emms-track-file-p track))
+      (emms-tag-editor--tagfile-function track)))
+
+(defun emms-tag-editor-edit-track (track &optional edit-anyway)
+  "Edit the track at point, or TRACK.
+If EDIT-ANYWAY is true or TRACK is not a file type, it will be loaded
+in the tag editor. Otherwise, if EMMS does not have a program configured
+to actually write tags to the audio file, do not open the tag data in
+the editor."
+  (interactive (list (emms-tag-editor-track-at)))
+  (cond
+   ((null track) (message "No track at point!"))
+   ((or (emms-tag-editor--track-editable-p track) edit-anyway)
+    (emms-tag-editor-insert-tracks (list track)))
+   (t (message "EMMS has no tag writing program configured for this file type!"))))
+
+(defun emms-tag-editor-edit-marked-tracks (&optional edit-anyway)
+  "Edit all tracks marked in the current buffer.
+If EDIT-ANYWAY is nil, filter out any file tracks that do not have a
+tagfile function defined."
   (interactive)
-  (let ((tracks (emms-mark-mapcar-marked-track 'emms-tag-editor-track-at t)))
+  (let* ((tracks (emms-mark-mapcar-marked-track 'emms-tag-editor-track-at t))
+         (funcs (mapcar #'emms-tag-editor--tagfile-function tracks)))
+    (when (seq-some #'null funcs)
+      (unless edit-anyway
+        (setq tracks (seq-filter #'emms-tag-editor--track-editable-p tracks))
+        (message "Skipped file tracks without a tag writing program configured.")))
     (if (null tracks)
-        (message "No track marked!")
+        (message "No writable track marked!")
       (emms-tag-editor-insert-tracks tracks))))
 
-(defun emms-tag-editor-edit ()
-  "Edit tags of either the track at point or all marked tracks."
-  (interactive)
+(defun emms-tag-editor-edit (&optional arg)
+  "Edit tags of either the track at point or all marked tracks.
+With a prefix argument, edits tags even if there is no external
+program for writing tags to the specified track or tracks."
+  (interactive "P")
   (if (emms-mark-has-markedp)
-      (emms-tag-editor-edit-marked-tracks)
-    (emms-tag-editor-edit-track (emms-tag-editor-track-at))))
+      (emms-tag-editor-edit-marked-tracks arg)
+    (emms-tag-editor-edit-track (emms-tag-editor-track-at) arg)))
 
 (defvar emms-tag-editor-mode-map
   (let ((map (make-sparse-keymap)))
-    (define-key map [tab] 'emms-tag-editor-next-field)
-    (define-key map [backtab] 'emms-tag-editor-prev-field)
-    (define-key map "\C-c\C-n" 'emms-tag-editor-next-track)
-    (define-key map "\C-c\C-p" 'emms-tag-editor-prev-track)
-    (define-key map "\C-c\C-c" 'emms-tag-editor-submit-and-exit)
-    (define-key map "\C-c\C-s" 'emms-tag-editor-submit)
-    (define-key map "\C-x\C-s" 'emms-tag-editor-submit)
-    (define-key map "\C-c\C-r" 'emms-tag-editor-set-all)
-    (define-key map "\C-c\C-a" 'emms-tag-editor-replace-in-tag)
-    (define-key map "\C-c\C-t" 'emms-tag-editor-transpose-tag)
+    ;; FIXME: Bind to "\t" rather than [tab] so it works in ttys as well.
+    (define-key map [tab] #'emms-tag-editor-next-field)
+    (define-key map [backtab] #'emms-tag-editor-prev-field)
+    (define-key map "\C-c\C-n" #'emms-tag-editor-next-track)
+    (define-key map "\C-c\C-p" #'emms-tag-editor-prev-track)
+    (define-key map "\C-c\C-c" #'emms-tag-editor-submit-and-exit)
+    (define-key map "\C-c\C-s" #'emms-tag-editor-submit)
+    (define-key map "\C-x\C-s" #'emms-tag-editor-submit)
+    (define-key map "\C-c\C-r" #'emms-tag-editor-set-all)
+    (define-key map "\C-c\C-a" #'emms-tag-editor-replace-in-tag)
+    (define-key map "\C-c\C-t" #'emms-tag-editor-transpose-tag)
     map)
   "Keymap for `emms-tag-editor-mode'.")
-(define-key emms-playlist-mode-map "E" 'emms-tag-editor-edit)
+(define-key emms-playlist-mode-map "E" #'emms-tag-editor-edit)
 
 (define-derived-mode emms-tag-editor-mode text-mode "Tag-Edit"
   "Major mode to edit track tags.
@@ -406,8 +436,8 @@ changes will only take effect on the tracks in the region."
    (let* ((tag1 (intern (emms-completing-read
                          "Tag 1: "
                          (mapcar (lambda (arg)
-                                         (list (symbol-name (car arg))))
-                                       emms-tag-editor-tags)
+                                   (list (symbol-name (car arg))))
+                                 emms-tag-editor-tags)
                          nil t)))
           (tag2 (intern (emms-completing-read
                          "Tag 2: "
@@ -454,25 +484,25 @@ C-u M-x emms-tag-editor-guess-tag-filename RET
     (read-from-minibuffer (format "Match in %sfile name(C-h for help): "
                                   (if current-prefix-arg "FULL " ""))
                           nil
-     (let ((map (make-sparse-keymap)))
-       (set-keymap-parent map minibuffer-local-map)
-       (define-key map "\C-h"
-         (lambda ()
-           (interactive)
-           (with-output-to-temp-buffer "*Help*"
-             (princ
-              "A pattern is a string like \"%a-%t-%y\" which stand for
+			  (let ((map (make-sparse-keymap)))
+			    (set-keymap-parent map minibuffer-local-map)
+			    (define-key map "\C-h"
+			      (lambda ()
+				(interactive)
+				(with-output-to-temp-buffer "*Help*"
+				  (princ
+				   "A pattern is a string like \"%a-%t-%y\" which stand for
 the file name is constructed by artist, title, year with seperator '-'.
 see `emms-tag-editor-compile-pattern' for detail about pattern syntax.
 
 Available tags are:
 ")
-             (mapc (lambda (tag)
-                     (princ (format "\t%s - %S\n" (cdr tag) (car tag))))
-                   emms-tag-editor-tags)
-             (with-current-buffer standard-output
-               (help-mode)))))
-       map))
+				  (mapc (lambda (tag)
+					  (princ (format "\t%s - %S\n" (cdr tag) (car tag))))
+					emms-tag-editor-tags)
+				  (with-current-buffer standard-output
+				    (help-mode)))))
+			    map))
     current-prefix-arg))
   (setq pattern (emms-tag-editor-compile-pattern pattern))
   (save-excursion
@@ -602,8 +632,9 @@ With prefix argument, bury the tag edit buffer."
     (if (not (and tracks (y-or-n-p "Submit changes? ")))
         (message "No tags were modified")
       (emms-tag-editor-erase-buffer emms-tag-editor-log-buffer)
-      (emms-tag-editor-apply tracks)))
-  (if arg (bury-buffer)))
+      (emms-tag-editor-apply tracks)
+      (emms-cache-save)))
+  (when arg (bury-buffer)))
 
 (defun emms-tag-editor-apply (tracks)
   "Apply all changes made to TRACKS."
@@ -616,7 +647,7 @@ With prefix argument, bury the tag edit buffer."
                 old (emms-track-get track 'orig-track))
           ;; rename local file
           (when (and (emms-track-get track 'newname)
-                     (eq (emms-track-get track 'type) 'file)
+                     (emms-track-file-p track)
                      (file-writable-p (emms-track-name track))
                      (y-or-n-p (format "Rename %s to %s? "
                                        (emms-track-name track)
@@ -636,19 +667,18 @@ With prefix argument, bury the tag edit buffer."
             (emms-track-set track 'name filename)
             (setq need-sync t)
             ;; register to emms-cache-db
-            (when (boundp 'emms-cache-modified-function)
+            (when (functionp emms-cache-modified-function)
               (funcall emms-cache-modified-function)
               (funcall emms-cache-set-function 'file filename old)))
           (emms-track-set track 'newname nil)
           ;; set tags to original track
           (dolist (tag emms-tag-editor-tags)
             (when (setq val (emms-track-get track (car tag)))
-            (emms-track-set old (car tag) val)))
+              (emms-track-set old (car tag) val)))
           ;; use external program to change tags in the file
-          (when (and (eq (emms-track-get track 'type) 'file)
+          (when (and (emms-track-file-p track)
                      (file-writable-p (emms-track-name track))
-                     (setq func (assoc (file-name-extension filename)
-                                       emms-tag-editor-tagfile-functions)))
+                     (setq func (emms-tag-editor--tagfile-function track)))
             (setq exit
                   (if (functionp (cdr func))
                       (funcall (cdr func) track)
@@ -666,11 +696,9 @@ With prefix argument, bury the tag edit buffer."
             (funcall emms-playlist-update-track-function))
           ;; clear modified tag
           (emms-track-set track 'tag-modified nil))))
-    (if (and (featurep 'emms-cache)
-             need-sync
-             (y-or-n-p "You have changed some track names; sync the cache? "))
-        (and (fboundp 'emms-cache-sync) ; silence byte-compiler
-             (emms-cache-sync)))
+    ;; sync the cache
+    (when need-sync
+      (emms-cache-sync nil))
     (unless (emms-tag-editor-display-log-buffer-maybe)
       (message "Setting tags...done"))))
 
@@ -725,7 +753,7 @@ With prefix argument, bury the tag edit buffer."
 (defun emms-tag-editor-log (&rest args)
   (with-current-buffer (get-buffer-create emms-tag-editor-log-buffer)
     (goto-char (point-max))
-    (insert (apply 'format args) "\n")))
+    (insert (apply #'format args) "\n")))
 
 ;;
 ;; Renaming files according their tags
@@ -754,7 +782,7 @@ value.
 If DONT-APPLY is non-nil the changes won't be applied directly.
 Then it's the callers job to apply them afterwards with
 `emms-tag-editor-apply'."
-  (if (eq (emms-track-get track 'type) 'file)
+  (if (emms-track-file-p track)
       (let* ((old-file (emms-track-name track))
              (path     (file-name-directory old-file))
              (suffix   (file-name-extension old-file))
@@ -762,15 +790,15 @@ Then it's the callers job to apply them afterwards with
                         path
                         (format-spec
                          emms-tag-editor-rename-format
-                         (apply 'format-spec-make
-                                (apply 'append
+                         (apply #'format-spec-make
+                                (apply #'append
                                        (mapcar
                                         (lambda (tag)
                                           (list (string-to-char (cdr tag))
                                                 (or (emms-track-get track (car tag))
                                                     "")))
                                         emms-tag-editor-tags))))
-                      "." suffix)))
+			"." suffix)))
         (emms-track-set track 'newname new-file)
         (emms-track-set track 'tag-modified t)
         (unless dont-apply
@@ -788,7 +816,7 @@ Then it's the callers job to apply them afterwards with
         (emms-tag-editor-rename-track track t))
       (emms-tag-editor-apply tracks))))
 
-(define-key emms-playlist-mode-map "R" 'emms-tag-editor-rename)
+(define-key emms-playlist-mode-map "R" #'emms-tag-editor-rename)
 
 (defvar emms-tag-editor-pipe-config
   '(("mid3iconv -e gbk <file>"
@@ -804,22 +832,22 @@ A pipe is defined like below:
 :arguments is a list or a function return list, for example:
 
   (\"--track-name\" name (\"--year\" info-year))
-  (lambda (track) (list (emms-track-name track 'name)))
+  (lambda (track) (list (emms-track-name track \\='name)))
 
-1. symbols can be 'name or elements of (mapcar 'car emms-tag-editor-tags),
+1. symbols can be \\='name or elements of (mapcar \\='car emms-tag-editor-tags),
    which will be replaced to track info before run command.
 2. sublist used to deal with group args, for example, (\"--year\" info-year), when
-   track's info-year is nil, the \"--year\" will be removed too.")
+   track\\='s info-year is nil, the \"--year\" will be removed too.")
 
 (defun emms-tag-editor-pipe-get (pipe-name key)
-  "Get the pipe value of KEY named PIPE-NAME in `emms-tag-editor-pipe-config'."
+  "Get the pipe value of KEY named PIPE-NAME in `emms-tag-editor-pipe-config\\='."
   (let ((config emms-tag-editor-pipe-config))
     (plist-get (cdr (assoc pipe-name config)) key)))
 
 (defun emms-tag-editor-pipe ()
   "Select and run pipe command to track at point or all marked tracks."
   (interactive)
-  (let* ((pipe-name (completing-read "Please choise pipe: " emms-tag-editor-pipe-config)))
+  (let* ((pipe-name (completing-read "Please choose pipe: " emms-tag-editor-pipe-config)))
     (when pipe-name
       (if (emms-mark-has-markedp)
           (emms-tag-editor-marked-track-pipe pipe-name)
@@ -828,7 +856,7 @@ A pipe is defined like below:
 
 (defun emms-tag-editor-track-pipe (track pipe-name)
   "Run command of pipe nameed PIPE-NAME to TRACK."
-  (if (eq (emms-track-get track 'type) 'file)
+  (if (emms-track-file-p track)
       (let* ((coding-system-for-read 'utf-8)
              (command (emms-tag-editor-pipe-get pipe-name :command))
              (arguments (emms-tag-editor-pipe-get pipe-name :arguments)))
@@ -876,6 +904,5 @@ A pipe is defined like below:
       (dolist (track tracks)
         (emms-tag-editor-track-pipe track pipe-name)))))
 
-
 (provide 'emms-tag-editor)
 ;;; Emms-tag-editor.el ends here
