@@ -1,10 +1,10 @@
 ;;; emms.el --- The Emacs Multimedia System  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2003-2022  Free Software Foundation, Inc.
+;; Copyright (C) 2003-2025  Free Software Foundation, Inc.
 
 ;; Author: Jorgen Schäfer <forcer@forcix.cx>, the Emms developers (see AUTHORS file)
 ;; Maintainer: Yoni Rabkin <yrk@gnu.org>
-;; Version: 20.2
+;; Version: 22
 ;; Keywords: emms, mp3, ogg, flac, music, mpeg, video, multimedia
 ;; Package-Type: multi
 ;; Package-Requires: ((cl-lib "0.5") (nadvice "0.3") (seq))
@@ -46,7 +46,7 @@
 (require 'emms-compat)
 (require 'seq)
 
-(defvar emms-version "20.2"
+(defvar emms-version "22"
   "EMMS version string.")
 
 ;;; User Customization
@@ -103,11 +103,9 @@ are played sequentially."
            (setq emms-player-next-function #'emms-next-noerror)))
   :type 'boolean)
 
-(defcustom emms-repeat-track nil
+(defvar-local emms-repeat-track nil
   "Non-nil, playback will repeat current track.  If nil, EMMS will play
-track by track normally."
-  :group 'emms
-  :type 'boolean)
+track by track normally.")
 
 (defvar-local emms-single-track nil
   "Non-nil, play the current track and then stop.")
@@ -587,7 +585,7 @@ See `emms-random-playlist'."
   (interactive)
   (customize-set-variable 'emms-random-playlist (not emms-random-playlist))
   (if emms-random-playlist
-	  (message "Will play the tracks randomly.")
+      (message "Will play the tracks randomly.")
     (message "Will play the tracks sequentially.")))
 
 (defun emms-toggle-repeat-playlist ()
@@ -831,38 +829,57 @@ for that purpose.")
   "Non-nil if the current buffer is an EMMS playlist.")
 (make-variable-buffer-local 'emms-playlist-buffer-p)
 
+(defvar emms-queue-lock nil
+  "The playlist name the active playlist queue is locked to, if any.")
+
 (defun emms-playlist-ensure-playlist-buffer ()
   "Throw an error if we're not in a playlist-buffer."
   (when (not emms-playlist-buffer-p)
     (error "Not an EMMS playlist buffer")))
 
+(defun emms-lock-queue ()
+  "Lock the current active playlist queue to it's playlist.
+This allows for uninterrupted playlist playback while browsing other playlists."
+  (interactive)
+  (setq emms-queue-lock (buffer-name emms-playlist-buffer))
+  (message (concat "Active playlist queue is locked to " emms-queue-lock)))
+
+(defun emms-unlock-queue ()
+  "Unlock the current active playlist queue from its playlist."
+  (interactive)
+  (setq emms-queue-lock nil)
+  (message "Active playlist queue is unlocked."))
+
 (defun emms-playlist-set-playlist-buffer (&optional buffer)
-  "Set the current playlist buffer."
+  "Set the current playlist buffer if the queue is not locked to it's playlist."
   (interactive
-   (list (let* ((buf-list (mapcar #'(lambda (buf)
-				      (list (buffer-name buf)))
-				  (emms-playlist-buffer-list)))
-		(sorted-buf-list (sort buf-list
-				       #'(lambda (lbuf rbuf)
-					   (< (length (car lbuf))
-					      (length (car rbuf))))))
-		(default (or (and emms-playlist-buffer-p
-				  ;; default to current buffer
-				  (buffer-name))
-			     ;; pick shortest buffer name, since it is
-			     ;; likely to be a shared prefix
-			     (car sorted-buf-list))))
-	   (emms-completing-read "Playlist buffer to make current: "
-				 sorted-buf-list nil t default))))
-  (let ((buf (if buffer
-		 (get-buffer buffer)
-	       (current-buffer))))
-    (with-current-buffer buf
-      (emms-playlist-ensure-playlist-buffer))
-    (setq emms-playlist-buffer buf)
-    (when (called-interactively-p 'interactive)
-      (message "Set current EMMS playlist buffer"))
-    buf))
+   (if (not emms-queue-lock)
+       (list (let* ((buf-list (mapcar #'(lambda (buf)
+				          (list (buffer-name buf)))
+				      (emms-playlist-buffer-list)))
+		    (sorted-buf-list (sort buf-list
+				           #'(lambda (lbuf rbuf)
+					       (< (length (car lbuf))
+					          (length (car rbuf))))))
+		    (default (or (and emms-playlist-buffer-p
+				      ;; default to current buffer
+				      (buffer-name))
+			         ;; pick shortest buffer name, since it is
+			         ;; likely to be a shared prefix
+			         (car sorted-buf-list))))
+	       (emms-completing-read "Playlist buffer to make current: "
+				     sorted-buf-list nil t default)))))
+  (if (not emms-queue-lock)
+      (let ((buf (if buffer
+		     (get-buffer buffer)
+	           (current-buffer))))
+        (with-current-buffer buf
+          (emms-playlist-ensure-playlist-buffer))
+        (setq emms-playlist-buffer buf)
+        (when (called-interactively-p 'interactive)
+          (message "Set current EMMS playlist buffer"))
+        buf)
+    (message (concat "The active playlist queue is locked to " emms-queue-lock))))
 
 (defun emms-playlist-new (&optional name)
   "Create a new playlist buffer.
@@ -1312,6 +1329,22 @@ ignore this."
   (ignore track) ;; explicit ignore
   t)
 
+(defun emms-playlist-length ()
+  "Display the total playing time of the current playlist."
+  (interactive)
+  (let ((acc 0))
+    (with-current-emms-playlist
+      (goto-char (point-min))
+      (emms-walk-tracks
+	(setq acc (+ acc (emms-track-get (emms-playlist-track-at (point)) 'info-playing-time)))))
+    (let ((minutes (/ acc 60))
+          (seconds (% acc 60)))
+      (message (if (>= minutes 60)
+		   (format "%d:%02d:%02d" (/ minutes 60) (% minutes 60) seconds)
+		 (format "%02d:%02d" minutes seconds))))))
+
+(defalias 'emms-playlist-total-playing-time 'emms-playlist-length)
+
 
 ;;; ------------------------------------------------------------------
 ;;; Helper functions
@@ -1435,6 +1468,7 @@ See emms-source-file.el for some examples."
 	(source-play (intern (format "emms-play-%s" name)))
 	(source-add (intern (format "emms-add-%s" name)))
 	(source-insert (intern (format "emms-insert-%s" name)))
+	(source-insert-current (intern (format "emms-insert-current-%s" name)))
 	(docstring "A source of tracks for EMMS.")
 	(interactive nil)
 	(call-args (delete '&rest
@@ -1467,7 +1501,11 @@ See emms-source-file.el for some examples."
        (defun ,source-insert ,arglist
 	 ,docstring
 	 ,interactive
-	 (emms-source-insert ',source-name ,@call-args)))))
+	 (emms-source-insert ',source-name ,@call-args))
+       (defun ,source-insert-current ,arglist
+	 ,docstring
+	 ,interactive
+	 (emms-source-insert-current ',source-name ,@call-args)))))
 
 (defun emms-source-play (source &rest args)
   "Play the tracks of SOURCE, after first clearing the EMMS playlist."
@@ -1489,8 +1527,16 @@ See emms-source-file.el for some examples."
 
 (defun emms-source-insert (source &rest args)
   "Insert the tracks from SOURCE in the current buffer."
-  (if (not emms-playlist-buffer-p)
-      (error "Not in an EMMS playlist buffer")
+  (with-current-emms-playlist
+    (apply #'emms-playlist-insert-source source args)))
+
+(defun emms-source-insert-current (source &rest args)
+  "Insert from SOURCE into buffer at current track."
+  (with-current-emms-playlist
+    (if (not emms-playlist-selected-marker)
+	(error "no current playlist track")
+      (goto-char emms-playlist-selected-marker)
+      (forward-line))
     (apply #'emms-playlist-insert-source source args)))
 
 ;;; User-defined playlists
